/*
 * Copyright 2021 Xilinx, Inc
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <x86_app.h>

bool IsSyslogEnabled(Sys* pSys)
{
    return pSys && pSys->yLogger;
}

bool IsVerboseEnabled(Sys* pSys)
{
    return pSys && pSys->yVerbose;
}

/*
 * --------------------------------------------------------------
 * Command line argument parsing
 * --------------------------------------------------------------
 */
bool ParseCmdLine(Sys* pSys,int argc,char* argv[])
{
    const struct option pLongOptions[] = {
        /*  name            has_arg             *flag       val */
        {   "logger",     no_argument,        0,          'l'},
        {   "verbose",    no_argument,        0,          'v'},
        {   "extra",      no_argument,        0,          'x'},
        {   "path",       required_argument,  0,          'p'},
        {   "images",     required_argument,  0,          'i'},
        {   "boards",     required_argument,  0,          'b'},
        {   "device",     required_argument,  0,          'd'},
        {   "commands",   required_argument,  0,          'c'},
        {   "help",       no_argument,        0,          'h'},
        {   0,              0,                  0,          0}
    };
    int iLongIndex = 0;
    int iChar;
    struct stat tStat;

    pSys->pcProgName = strrchr(argv[0],'/');
    if(pSys->pcProgName) {
        pSys->pcProgName++;
    } else {
        pSys->pcProgName = argv[0];
    }

    for(
            ;
            (
                iLongIndex = 0,
                iChar = getopt_long(argc,argv,":lvxp:i:b:d:c:h",pLongOptions,&iLongIndex),
                iChar
            ) >= 0;
       ) {
        switch(iChar) {
        default:    /* FallThrough */
        case '?':
            /* unknown option */
            if(0 == optopt) {
                Die(0,0,"Unknown option '%s'",argv[optind - 1]);
            } else {
                Die(0,0,"Unknown option '-%c'",optopt);
            }
            break;

        case ':':
            /* Missing mandatory argument */
            if(0 == optopt) {
                Die(0,0,"Option '%s' missing mandatory argument",argv[optind - 1]);
            } else {
                Die(0,0,"Option '-%c' missing mandatory argument",optopt);
            }
            break;

        case 'l':
            pSys->yLogger = true;
            break;

        case 'v':
            pSys->yVerbose = true;
            break;

        case 'x':
            pSys->yExtra = true;
            break;

        case 'p':
            if(pSys->pcImagesPath) {
                Die(0,0,"Images path specified multiple times");
                /* Not reached */
                return false;
            }
            pSys->pcImagesPath = optarg;
            if(-1 == stat(pSys->pcImagesPath,&tStat)) {
                Die(0,errno,"Invalid path %s",pSys->pcImagesPath);
                /* Not reached */
                return false;
            } else if(!S_ISDIR(tStat.st_mode)) {
                Die(0,errno,"Path %s is not a directory",pSys->pcImagesPath);
                /* Not reached */
                return false;
            }
            pSys->qImagesPathLen = strlen(pSys->pcImagesPath);
            break;

        case 'i':
            if(pSys->pcImagesCsv) {
                Die(0,0,"Images CSV specified multiple times");
                /* Not reached */
                return false;
            }
            pSys->pcImagesCsv = optarg;
            if(-1 == stat(pSys->pcImagesCsv,&tStat)) {
                Die(0,errno,"Invalid Images CSV file '%s'",pSys->pcImagesCsv);
                /* Not reached */
                return false;
            } else if(!S_ISREG(tStat.st_mode)) {
                Die(0,errno,"Images CSV file '%s' is not a regular file",pSys->pcImagesCsv);
                /* Not reached */
                return false;
            }
            InitListHead(&pSys->tImagesCsvListHead,pSys,pSys->pcImagesCsv);
            break;

        case 'b':
            if(pSys->pcBoardsCsv) {
                Die(0,0,"Boards CSV specified multiple times");
                /* Not reached */
                return false;
            }
            pSys->pcBoardsCsv = optarg;
            if(-1 == stat(pSys->pcBoardsCsv,&tStat)) {
                Die(0,errno,"Invalid Boards CSV file '%s'",pSys->pcBoardsCsv);
                /* Not reached */
                return false;
            } else if(!S_ISREG(tStat.st_mode)) {
                Die(0,errno,"Boards CSV file '%s' is not a regular file",pSys->pcBoardsCsv);
                /* Not reached */
                return false;
            }
            InitListHead(&pSys->tBoardsCsvListHead,pSys,pSys->pcBoardsCsv);
            break;

        case 'd':
            if(pSys->pcDevice) {
                Die(0,0,"Device specified multiple times");
                /* Not reached */
                return false;
            }
            pSys->pcDevice = optarg;
            break;

        case 'c':
            if(pSys->pcCommandFile) {
                Die(0,0,"CommandFile specified multiple times");
                /* Not reached */
                return false;
            }
            pSys->pcCommandFile = optarg;
            if(-1 == stat(pSys->pcCommandFile,&tStat)) {
                Die(0,errno,"Invalid Command file path %s",pSys->pcCommandFile);
                /* Not reached */
                return false;
            } else if(!S_ISREG(tStat.st_mode)) {
                Die(0,errno,"Command file %s is not a regular file",pSys->pcCommandFile);
                /* Not reached */
                return false;
            }
            break;

        case 'h':
            Help(0);
            return false;
        }
    }

    if(!pSys->pcImagesPath) {
        Die(0,0,"Images path not specified");
    }
    if(!pSys->pcImagesCsv) {
        Die(0,0,"Images CSV file not specified");
    }
    if(!pSys->pcBoardsCsv) {
        Die(0,0,"Boards CSV file not specified");
    }

    if(pSys->pcCommandFile) {
        pSys->pCommandFile = fopen(pSys->pcCommandFile,"r");
        if(!pSys->pCommandFile) {
            Die(0,errno,"%s",pSys->pcCommandFile);
            /* Not reached */
            return false;
        }
    } else {
        pSys->pcCommandFile = "stdin";
        pSys->pCommandFile = stdin;
    }

    if(isatty(fileno(pSys->pCommandFile)) && pSys->yLogger) {
        pSys->yLogger = false;
        Warn(0,0,"Disabling use of system logger because commands are interactive");
    }

    return true;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */

bool InitSys(Sys** ppSys,int argc,char* argv[])
{
    Sys* pSys;

    pSys = calloc(1,sizeof(Sys));
    if(!pSys) {
        fprintf(stderr,"calloc(Sys) -> %s(%d)\n",strerror(errno),errno);
        /* Not reached */
        return false;
    }
    *ppSys = pSys;

    if(!ParseCmdLine(pSys,argc,argv)) {
        return false;
    }

    if(pSys->yLogger) {
        openlog(pSys->pcProgName,LOG_PID|LOG_NDELAY,LOG_USER);
    }

    if(!ParseBoardsCsv(pSys,&pSys->tBoardsCsvListHead)) {
        return false;
    }

    if(!ParseImagesCsv(pSys,&pSys->tImagesCsvListHead)) {
        return false;
    }

    if(pSys->pcDevice) {
        if(!OpenDevice(pSys,"CommandLine",0,pSys->pcDevice)) {
            /* Not reached */
            return false;
        }
    }

    return true;
}

void CloseSys(Sys* pSys)
{
    if(pSys->pCommandFile && (pSys->pCommandFile != stdin)) {
        fclose(pSys->pCommandFile);
        pSys->pCommandFile = 0;
    }
    FreeList(&pSys->tBoardsCsvListHead);
    FreeList(&pSys->tImagesCsvListHead);
    closelog();
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */

bool DoDevice(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
    if(!strcasecmp(ppcTokens[0],"device")) {
        if(2 == iNumTokens) {
            if(pSys->pcDevice) {
                CloseDevice(pSys);
            }
            return OpenDevice(pSys,pSys->pcCommandFile,pSys->iCommandLine,ppcTokens[1]);
        } else {
            Warn(pSys,0,"-------- Syntax error --------");
        }
    } else if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,
                "%-24sSwitch to specified board type and BDF and run DevInfo",
                "Device BoardType:<BDF>"
                );
    }
    return false;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */

bool DoDevInfo(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
    if(!strcasecmp(ppcTokens[0],"DevInfo")) {
        if(1 == iNumTokens) {
            (void)DecodeDevInfo(pSys);
        } else {
            Warn(pSys,0,"-------- Syntax error --------");
        }
        return true;
    } else if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,"%-24sPrint device info: decode multiboot,\n%24slist and verify QSPI images","DevInfo","");
    }
    return false;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */
bool FindAndListImage(ListHead* pListHead,ListElem* pListElem,void* pxArg)
{
    ImageInfo* pII = (ImageInfo*)pListElem;
    int iIndex = *(int*)pxArg;

    if(0 == iIndex) {
        Info(pListHead->pSys,"ImagesPath: %s",pListHead->pSys->pcImagesPath);
    }
    Info(pListHead->pSys,"#%d: %s",iIndex,pII->tCR.ppcTokens[0]);
    switch(pII->eType) {
    default: Info(pListHead->pSys,"\tType: ???"); break;
    case IMAGE_TYPE_PROM: Info(pListHead->pSys,"\tType: TANDEM_PROM"); break;
    case IMAGE_TYPE_STAGE1: Info(pListHead->pSys,"\tType: TANDEM_PCIE:STAGE1"); break;
    case IMAGE_TYPE_STAGE2: Info(pListHead->pSys,"\tType: TANDEM_PCIE:STAGE2"); break;
    }
    Info(pListHead->pSys,"\tChecksum: %08x",pII->qChecksum);
    Info(pListHead->pSys,"\tId: %s",pII->tCR.ppcTokens[2]);

    *(int*)pxArg = ++iIndex;
    return false;
}

bool DoList(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
    if(!strcasecmp(ppcTokens[0],"List")) {
        if(1 == iNumTokens) {
            int i = 0;
            (void)CheckElems(&pSys->tImagesCsvListHead,FindAndListImage,&i);
            return true;
        } else {
            Warn(pSys,0,"-------- Syntax error --------");
        }
    } else if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,"%-24sList all files in database,\n%24swith a unique index for programming","List","");
    }
    return false;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */

static bool FindImage(ListHead* pListHead,ListElem* pListElem,void* pxArg)
{
    uint32_t qImageIndex = *(uint32_t*)pxArg;

    if(qImageIndex > 0) {
        *(uint32_t*)pxArg = qImageIndex - 1;
        return false;
    }
    return true;
}

bool DoProg(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
    ImageInfo* pII;
    ReservedBlockEntry* pRBE;

    if(!strcasecmp(ppcTokens[0],"Prog")) {
        if(2 == iNumTokens) {
            uint32_t qImageIndex;

            if(!GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                "qImageIndex",ppcTokens[1],10,&qImageIndex)) {
                Warn(pSys,0,"Index '%s' is not a valid number",ppcTokens[1]);
                return true;
            }
            pII = (ImageInfo*) CheckElems(&pSys->tImagesCsvListHead,FindImage,&qImageIndex);
            if(!pII) {
                Warn(pSys,0,"Index '%s' is out of range",ppcTokens[1]);
                return true;
            }
            pRBE = pSys->ppRBE[0];
            if(ERR_NONE == pRBE->qErrCode) {
                Info(pSys,"Programming A after copying A to B");
                if(!EraseLoc(pSys,'b')) {
                    Warn(pSys,0,"Could not erase image and/or reserved entry for B");
                    return true;
                }
                if(!CopyAtoB(pSys)) {
                    Warn(pSys,0,"Could not copy image and/or reserved entry from A to B");
                    return true;
                }
                if(!EraseLoc(pSys,'a')) {
                    Warn(pSys,0,"Could not erase image and/or reserved entry for A");
                    return true;
                }
                if(!ProgramImageAtLoc(pSys,pII,'a')) {
                    Warn(pSys,0,"Could not program image and/or reserved entry for A");
                    return true;
                }
            } else {
                Info(pSys,"No good image at A; programming both A and B");
                if(!EraseLoc(pSys,'a')) {
                    Warn(pSys,0,"Could not erase image and/or reserved entry for A");
                    return true;
                }
                if(!EraseLoc(pSys,'b')) {
                    Warn(pSys,0,"Could not erase image and/or reserved entry for B");
                    return true;
                }
                if(!ProgramImageAtLoc(pSys,pII,'a')) {
                    Warn(pSys,0,"Could not program image and/or reserved entry for A");
                    return true;
                }
                if(!ProgramImageAtLoc(pSys,pII,'b')) {
                    Warn(pSys,0,"Could not program image and/or reserved entry for B");
                    return true;
                }
            }
            return true;
        } else {
            Warn(pSys,0,"-------- Syntax error --------");
        }
    } else if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,
            "%-24sErase B\n%24sCopy A to B\n%24sErase A\n%24sProgram file[Index] into position A\n%24sUpdate R",
            "Prog <Index>",
            "",
            "",
            "",
            ""
            );
    }
    return false;
}

bool DoProgLoc(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
    ImageInfo* pII;

    if(!strcasecmp(ppcTokens[0],"ProgLoc")) {
        if(3 == iNumTokens) {
            uint32_t qImageIndex;

            if(!GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                "qImageIndex",ppcTokens[1],10,&qImageIndex)) {
                Warn(pSys,0,"Index '%s' is not a valid number",ppcTokens[1]);
                return true;
            }
            pII = (ImageInfo*) CheckElems(&pSys->tImagesCsvListHead,FindImage,&qImageIndex);
            if(!pII) {
                Warn(pSys,0,"Index '%s' is out of range",ppcTokens[1]);
                return true;
            }
            if(1
                    && (0 != strcasecmp(ppcTokens[2],"a"))
                    && (0 != strcasecmp(ppcTokens[2],"b"))
                    && (0 != strcasecmp(ppcTokens[2],"g"))
                ) {
                Warn(pSys,0,"Loc '%s' must be one of a, b or g",ppcTokens[2]);
                return true;
            }
            if(!EraseLoc(pSys,(uint32_t)tolower(ppcTokens[2][0]))) {
                Warn(pSys,0,"Could not erase image and/or reserved entry for '%c'",tolower(ppcTokens[2][0]));
                return true;
            }
            if(!ProgramImageAtLoc(pSys,pII,(uint32_t)tolower(ppcTokens[2][0]))) {
                Warn(pSys,0,"Could not program image and/or reserved entry for '%c'",tolower(ppcTokens[2][0]));
                return true;
            }
            return true;
        } else {
            Warn(pSys,0,"-------- Syntax error --------");
        }
    } else if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,
                "%-24sErase <Loc>\n%24sProgram file[Index] into position A\n%24sUpdate R",
                "ProgLoc <Index> <Loc>",
                "",
                ""
                );
    }
    return false;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */

bool DoExit(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
    if(0
            || !strcasecmp(ppcTokens[0],"Exit")
            || !strcasecmp(ppcTokens[0],"Quit")
            ) {
        if(1 == iNumTokens) {
            pSys->yStop = true;
        } else {
            Warn(pSys,0,"-------- Syntax error --------");
        }
        return true;
    } else if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,"%-24sExit program","Exit");
        Info(pSys,"%-24sExit program","Quit");
    }
    return false;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */

bool DoVerbose(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
    if(!strcasecmp(ppcTokens[0],"Verbose")) {
        if(1 == iNumTokens) {
            pSys->yVerbose ^= pSys->yVerbose;
        } else {
            Warn(pSys,0,"-------- Syntax error --------");
        }
        return true;
    } else if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,"%-24sToggle verbosity","Verbose");
    }
    return false;
}
/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */

bool DoExtra(Sys* pSys,int iNumTokens,char* ppcTokens[])
{
#define ENTRY(CMD,help)    {COMMAND_ ## CMD, #CMD, help}
    CmdInfo g_pExtraCmdInfo[] = {
        ENTRY(VERIFY_FLASH, "FLASH_POSITION FLASH_MAX_SPEED_HZ PEB_SIZE_BYTES NUM_PEBS SUBSECTOR_SIZE_BYTES"),
        ENTRY(VERIFY_IMAGE,         "IMAGE_OFFSET IMAGE_SIZE IMAGE_CHECKSUM"),
        ENTRY(GET_MULTIBOOT_BYTES,  ""),
        ENTRY(ERASE_PEBS,           "PEB_START PEB_END"),
        ENTRY(READ_PEB,             "PEB_NUMBER FILENAME"),
        ENTRY(WRITE_PEB,            "PEB_NUMBER FILENAME"),
        ENTRY(ERASE_SUBSECTOR,      "SUBSECTOR_NUMBER"),
        ENTRY(READ_SUBSECTOR,       "SUBSECTOR_NUMBER FILENAME"),
        ENTRY(WRITE_SUBSECTOR,      "SUBSECTOR_NUMBER FILENAME"),
        ENTRY(SIM_STOP,             ""),
        ENTRY(INVALID,              "")
    };
    CmdInfo* pCI;
    uint32_t qErrCode;
    uint32_t qErrBSP;

    if(!strcasecmp(ppcTokens[0],"help")) {
        Info(pSys,"---- Extra commands ----");
        for(pCI = g_pExtraCmdInfo; pCI->eCommand != COMMAND_INVALID; pCI++) {
            Info(pSys,"%-24s%s",pCI->pcCmd,pCI->pcHelp);
        }
        return false;
    }
    if(!pSys || !pSys->yExtra) {
        return false;
    } else if(!pSys->pBI) {
        Warn(pSys,0,"Board not defined");
        return false;
    }

    for(pCI = g_pExtraCmdInfo; pCI->eCommand != COMMAND_INVALID; pCI++) {
        if(!strcasecmp(ppcTokens[0],pCI->pcCmd)) {
            qErrCode = 0;
            switch(pCI->eCommand) {
            default:        /* FallThrough */
            case COMMAND_INVALID:
                return false;

            case COMMAND_VERIFY_FLASH:
                if(1 == iNumTokens) {
                    qErrCode = VerifyFlash(
                            pSys,
                            pSys->pBI->qConnectionMode,
                            pSys->pBI->qBusWidth,
                            pSys->pBI->qQspiPeripheralInputSpeedHz,
                            pSys->pBI->qPrescaler,
                            pSys->pBI->qCacheCoherentDMA,
                            pSys->pBI->qPEBSize,
                            pSys->pBI->qNumPEBs,
                            pSys->pBI->qSubsectorSize,
                            &qErrBSP
                            );
                } else {
                    Warn(pSys,0,"Syntax error");
                }
                break;

            case COMMAND_VERIFY_IMAGE:
                {
                    uint32_t qImageOffset;
                    uint32_t qImageSize;
                    uint32_t qImageChecksum;

                    if(1
                            && (4 == iNumTokens)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qImageOffset",ppcTokens[1],10,&qImageOffset)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qImageSize",ppcTokens[2],10,&qImageSize)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qImageChecksum",ppcTokens[3],10,&qImageChecksum)
                        ) {
                        qErrCode = VerifyImage(pSys,qImageOffset,qImageSize,qImageChecksum,&qErrBSP);
                    } else {
                        Warn(pSys,0,"Syntax error");
                    }
                }
                break;

            case COMMAND_GET_MULTIBOOT_BYTES:
                {
                    uint32_t qMultibootBytes;

                    qErrCode = GetMultibootBytes(pSys,&qErrBSP,&qMultibootBytes);
                    Info(pSys,"MultibootBytes = 0x%08x",qMultibootBytes);
                }
                break;

            case COMMAND_ERASE_PEBS:
                {
                    uint32_t qPEBStart;
                    uint32_t qPEBEnd;
                    uint32_t qErrBSP;

                    if(1
                            && (3 == iNumTokens)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qPEBStart",ppcTokens[1],10,&qPEBStart)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qPEBEnd",ppcTokens[2],10,&qPEBEnd)
                        ) {
                        qErrCode = ErasePEBs(pSys,qPEBStart,qPEBEnd,&qErrBSP);
                    } else {
                        Warn(pSys,0,"Syntax error");
                    }
                }
                break;

            case COMMAND_READ_PEB:
                {
                    uint32_t qPEBNumber;
                    void* pxMap;
                    uint32_t qTmp;

                    if(1
                            && (3 == iNumTokens)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qPEBNumber",ppcTokens[1],10,&qPEBNumber)
                        ) {
                        qTmp = pSys->pBI->qPEBSize;
                        if(MapFile(
                                pSys,                       /* Sys* pSys, */
                                pCI->pcCmd,                 /* const char* pcContext, */
                                ppcTokens[2],               /* const char* pcFile, */
                                MAPPING_W_FIXED_MAY_CREATE, /* MAPPING eMapping, */
                                0,                          /* uint64_t gOffset */
                                &qTmp,                      /* uint32_t* pqRequiredSize, */
                                &pxMap                      /* void** ppxMap */
                            )) {
                            qErrCode = ReadPEB(pSys,qPEBNumber,&qErrBSP,(uint8_t*)pxMap);
                            (void)UnmapFile(pSys,pSys->pBI->qPEBSize,pxMap);
                        }
                    }
                }
                break;

            case COMMAND_WRITE_PEB:
                {
                    uint32_t qPEBNumber;
                    void* pxMap;
                    uint32_t qTmp;

                    if(1
                            && (3 == iNumTokens)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qPEBNumber",ppcTokens[1],10,&qPEBNumber)
                        ) {
                        qTmp = pSys->pBI->qPEBSize;
                        if(MapFile(
                                pSys,               /* Sys* pSys, */
                                pCI->pcCmd,         /* const char* pcContext, */
                                ppcTokens[2],       /* const char* pcFile, */
                                MAPPING_R_FIXED,    /* MAPPING eMapping, */
                                0,                  /* uint64_t gOffset */
                                &qTmp,              /* uint32_t* pqRequiredSize, */
                                &pxMap              /* void** ppxMap */
                            )) {
                            qErrCode = WritePEB(pSys,qPEBNumber,(uint8_t*)pxMap,&qErrBSP);
                            (void)UnmapFile(pSys,pSys->pBI->qPEBSize,pxMap);
                        }
                    }
                }
                break;

            case COMMAND_ERASE_SUBSECTOR:
                {
                    uint32_t qSubsectorNumber;

                    if(1
                            && (2 == iNumTokens)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qSubsectorNumber",ppcTokens[1],10,&qSubsectorNumber)
                        ) {
                        qErrCode = EraseSubsector(pSys,qSubsectorNumber,&qErrBSP);
                    }
                }
                break;

            case COMMAND_READ_SUBSECTOR:
                {
                    uint32_t qSubsectorNumber;
                    void* pxMap;
                    uint32_t qTmp;

                    if(1
                            && (3 == iNumTokens)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qSubsectorNumber",ppcTokens[1],10,&qSubsectorNumber)
                        ) {
                        qTmp = pSys->pBI->qSubsectorSize;
                        if(MapFile(
                                pSys,                       /* Sys* pSys, */
                                pCI->pcCmd,                 /* const char* pcContext, */
                                ppcTokens[2],               /* const char* pcFile, */
                                MAPPING_W_FIXED_MAY_CREATE, /* MAPPING eMapping, */
                                0,                          /* uint64_t gOffset */
                                &qTmp,                      /* uint32_t qRequiredSize, */
                                &pxMap                      /* void** ppxMap */
                            )) {
                            qErrCode = ReadSubsector(pSys,qSubsectorNumber,&qErrBSP,(uint8_t*)pxMap);
                            (void)UnmapFile(pSys,pSys->pBI->qSubsectorSize,pxMap);
                        }
                    }
                }
                break;

            case COMMAND_WRITE_SUBSECTOR:
                {
                    uint32_t qSubsectorNumber;
                    void* pxMap;
                    uint32_t qTmp;

                    if(1
                            && (3 == iNumTokens)
                            && GetBaseNumber(pSys,pSys->pcCommandFile,pSys->iCommandLine,
                                                "qSubsectorNumber",ppcTokens[1],10,&qSubsectorNumber)
                        ) {
                        qTmp = pSys->pBI->qSubsectorSize;
                        if(MapFile(
                                pSys,                       /* Sys* pSys, */
                                pCI->pcCmd,                 /* const char* pcContext, */
                                ppcTokens[2],               /* const char* pcFile, */
                                MAPPING_R_FIXED,            /* MAPPING eMapping, */
                                0,                          /* uint64_t gOffset */
                                &qTmp,                      /* uint32_t qRequiredSize, */
                                &pxMap                      /* void** ppxMap */
                            )) {
                            qErrCode = WriteSubsector(pSys,qSubsectorNumber,(uint8_t*)pxMap,&qErrBSP);
                            (void)UnmapFile(pSys,pSys->pBI->qSubsectorSize,pxMap);
                        }
                    }
                }
                break;

            case COMMAND_SIM_STOP:
                qErrCode = SimStop(pSys,&qErrBSP);
                break;
            }
            Info(pSys,"\t-> %08x (%s) (ErrBSP %d)",qErrCode,GetErrString((ERR)qErrCode),qErrBSP);
            return true;
        }
    }
    return false;
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */
typedef bool (*Handler)(Sys* pSys,int iNumTokens,char*ppcTokens[]);
const Handler pHandlers[] = {
    DoDevice,
    DoDevInfo,
    DoList,
    DoProg,
    DoProgLoc,
    DoExit,
    DoVerbose,
    DoExtra,
    0
};

void Help(Sys* pSys)
{
    const Handler* pHandler;
    char* ppcHelpTokens[] = {"help",0};

    for(pHandler = pHandlers; *pHandler; pHandler++) {
        (*pHandler)(pSys,1,ppcHelpTokens);
    }
}

/*
 * --------------------------------------------------------------
 * --------------------------------------------------------------
 */
#define MAX_LINE    256
#define MAX_INPUT_TOKENS    16
int main(int argc,char* argv[])
{
    char pcLine[MAX_LINE];
    char* ppcTokens[MAX_INPUT_TOKENS];
    int iNumTokens;
    Sys* pSys;
    const Handler* pHandler;
    bool yHandled;

    if(!InitSys(&pSys,argc,argv)) {
        /* Not reached */
        exit(-1);
    }

    pSys->yStop = false;
    for(pSys->iCommandLine = 0; !pSys->yStop; pSys->iCommandLine++) {
        if(isatty(fileno(pSys->pCommandFile))) {
            printf("%s> ",pSys->pcProgName); fflush(stdout);
        }
        if(!fgets(pcLine,MAX_LINE,pSys->pCommandFile)) {
            printf("\n");
            break;
        }
        pcLine[MAX_LINE - 1] = 0;
        iNumTokens = Tokenize(
                pSys,
                pSys->pcCommandFile,
                pSys->iCommandLine,
                pcLine,
                " \t\n\r",
                MAX_INPUT_TOKENS,
                ppcTokens
                );
        if(iNumTokens <= 0) {
            continue;
        }

        yHandled = 0 == strcasecmp(ppcTokens[0],"help");
        for(pHandler = pHandlers; *pHandler; pHandler++) {
            if((*pHandler)(pSys,iNumTokens,ppcTokens)) {
                yHandled = true;
                break;
            }
        }
        if(!yHandled) {
            Help(pSys);
        }
    }
    CloseSys(pSys);
    return 0;
}
